import abc
import json
import os
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import Literal

import numpy as np
import polars as pl
from sklearn.preprocessing import KBinsDiscretizer

from replay.utils import (
    PYSPARK_AVAILABLE,
    DataFrameLike,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)

if PYSPARK_AVAILABLE:  # pragma: no cover
    from pyspark.ml.feature import Bucketizer, QuantileDiscretizer
    from pyspark.sql.functions import isnan

HandleInvalidStrategies = Literal["error", "skip", "keep"]


class BaseDiscretizingRule(abc.ABC):  # pragma: no cover
    """
    Interface of the discretizing rule
    """

    @property
    @abc.abstractmethod
    def column(self) -> str:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def n_bins(self) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def fit(self, df: DataFrameLike) -> "BaseDiscretizingRule":
        raise NotImplementedError()

    @abc.abstractmethod
    def partial_fit(self, df: DataFrameLike) -> "BaseDiscretizingRule":
        raise NotImplementedError()

    @abc.abstractmethod
    def transform(self, df: DataFrameLike) -> DataFrameLike:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_handle_invalid(self, handle_invalid: HandleInvalidStrategies) -> None:
        raise NotImplementedError()

    def fit_transform(self, df: DataFrameLike) -> DataFrameLike:
        return self.fit(df).transform(df)


class GreedyDiscretizingRule(BaseDiscretizingRule):
    """
    Implementation of the Discretizing rule for a column of PySpark, Polars and Pandas DataFrames.
    Discretizes column values according to the Greedy binning strategy:
    https://github.com/microsoft/LightGBM/blob/master/src/io/bin.cpp#L78
    It is recommended to use together with the Discretizer.
    """

    _HANDLE_INVALID_STRATEGIES = ("error", "skip", "keep")

    def __init__(
        self,
        column: str,
        n_bins: int,
        min_data_in_bin: int = 1,
        handle_invalid: HandleInvalidStrategies = "keep",
    ) -> None:
        """
        :param column: Name of the column to discretize.
        :param n_bins: Number of intervals where data will be binned.
        :param min_data_in_bin: Minimum number of samples in one bin.
        :param handle_invalid: handle_invalid rule.
            indicates how to process NaN in data.
            If ``skip`` - filter out rows with invalid values.
            If ``error`` - throw an error.
            If ``keep`` - keep invalid values in a special additional bucket with number = n_bins.
            Default ``keep``.
        """
        self._n_bins = n_bins
        self._col = column
        self._min_data_in_bin = min_data_in_bin
        self._bins = None
        self._is_fitted = False

        if handle_invalid not in self._HANDLE_INVALID_STRATEGIES:
            msg = f"handle_invalid should be either 'error' or 'skip' or 'keep', got {handle_invalid}."
            raise ValueError(msg)
        self._handle_invalid = handle_invalid

    @property
    def column(self) -> str:
        return self._col

    @property
    def n_bins(self) -> str:
        return self._n_bins

    def _greedy_bin_find(
        self,
        distinct_values: np.ndarray,
        counts: np.ndarray,
        num_distinct_values: int,
        max_bin: int,
        total_cnt: int,
        min_data_in_bin: int,
    ) -> list[float]:
        """
        Computes bound for bins.

        :param distinct_values: Array of unique values.
        :param counts: Number of samples corresponding to the every unique value.
        :param num_distinct_values: Number of unique value.
        :param max_bin: Maximum bin number.
        :param total_cnt: Total number of samples.
        :param min_data_in_bin: Minimum number of samples in one bin.
        """
        bin_upper_bound = []
        assert max_bin > 0

        if total_cnt < max_bin * min_data_in_bin:
            warn_msg = f"Expected at least {max_bin*min_data_in_bin} samples (n_bins*min_data_in_bin) \
= ({self._n_bins}*{min_data_in_bin}). Got {total_cnt}. The number of bins will be less in the result"
            warnings.warn(warn_msg)
        if num_distinct_values <= max_bin:
            cur_cnt_inbin = 0
            for i in range(num_distinct_values - 1):
                cur_cnt_inbin += counts[i]
                if cur_cnt_inbin >= min_data_in_bin:
                    bin_upper_bound.append((distinct_values[i] + distinct_values[i + 1]) / 2.0)
                    cur_cnt_inbin = 0

            cur_cnt_inbin += counts[num_distinct_values - 1]
            bin_upper_bound.append(float("Inf"))

        else:
            if min_data_in_bin > 0:
                max_bin = min(max_bin, total_cnt // min_data_in_bin)
                max_bin = max(max_bin, 1)
            mean_bin_size = total_cnt / max_bin
            rest_bin_cnt = max_bin
            rest_sample_cnt = total_cnt

            is_big_count_value = counts >= mean_bin_size
            rest_bin_cnt -= np.sum(is_big_count_value)
            rest_sample_cnt -= np.sum(counts[is_big_count_value])

            mean_bin_size = rest_sample_cnt / rest_bin_cnt
            upper_bounds = [float("Inf")] * max_bin
            lower_bounds = [float("Inf")] * max_bin

            bin_cnt = 0
            lower_bounds[bin_cnt] = distinct_values[0]
            cur_cnt_inbin = 0

            for i in range(num_distinct_values - 1):  # pragma: no cover
                if not is_big_count_value[i]:
                    rest_sample_cnt -= counts[i]

                cur_cnt_inbin += counts[i]

                if (
                    is_big_count_value[i]
                    or cur_cnt_inbin >= mean_bin_size
                    or (is_big_count_value[i + 1] and cur_cnt_inbin >= max(1.0, mean_bin_size * 0.5))
                ):
                    upper_bounds[bin_cnt] = distinct_values[i]
                    bin_cnt += 1
                    lower_bounds[bin_cnt] = distinct_values[i + 1]
                    if bin_cnt >= max_bin - 1:
                        break
                    cur_cnt_inbin = 0
                    if not is_big_count_value[i]:
                        rest_bin_cnt -= 1
                        mean_bin_size = rest_sample_cnt / rest_bin_cnt

            bin_upper_bound = [(upper_bounds[i] + lower_bounds[i + 1]) / 2.0 for i in range(bin_cnt - 1)]
            bin_upper_bound.append(float("Inf"))
        return bin_upper_bound

    def _fit_spark(self, df: SparkDataFrame) -> None:
        warn_msg = "DataFrame will be partially converted to the Pandas type during internal calculations in 'fit'"
        warnings.warn(warn_msg)
        value_counts = df.groupBy(self._col).count().orderBy(self._col).toPandas()
        bins = [-float("inf")]
        bins += self._greedy_bin_find(
            value_counts[self._col].values,
            value_counts["count"].values,
            value_counts.shape[0],
            self._n_bins + 1,
            df.count(),
            self._min_data_in_bin,
        )
        self._bins = bins

    def _fit_pandas(self, df: PandasDataFrame) -> None:
        vc = df[self._col].value_counts().sort_index()
        bins = self._greedy_bin_find(
            vc.index.values, vc.values, len(vc), self._n_bins + 1, vc.sum(), self._min_data_in_bin
        )
        self._bins = [-np.inf, *bins]

    def _fit_polars(self, df: PolarsDataFrame) -> None:
        warn_msg = "DataFrame will be converted to the Pandas type during internal calculations in 'fit'"
        warnings.warn(warn_msg)
        self._fit_pandas(df.to_pandas())

    def fit(self, df: DataFrameLike) -> "GreedyDiscretizingRule":
        """
        Fits Discretizing Rule to input dataframe.

        :param df: input dataframe.
        :returns: fitted DiscretizingRule.
        """
        if self._is_fitted:
            return self

        df = self._validate_input(df)

        if isinstance(df, PandasDataFrame):
            self._fit_pandas(df)
        elif isinstance(df, SparkDataFrame):
            self._fit_spark(df)
        else:
            self._fit_polars(df)

        self._is_fitted = True
        return self

    def partial_fit(self, df: DataFrameLike) -> "GreedyDiscretizingRule":
        """
        Fits new data to already fitted DiscretizingRule.

        :param df: input dataframe.
        :returns: fitted DiscretizingRule.
        """
        if not self._is_fitted:
            return self.fit(df)

        msg = f"{self.__class__.__name__} is not implemented for partial_fit yet."
        raise NotImplementedError(msg)

    def _transform_pandas(self, df: PandasDataFrame) -> PandasDataFrame:
        binned_column = np.digitize(df[self._col].values, self._bins)
        binned_column -= 1
        df_transformed = df.copy()
        df_transformed.loc[:, self._col] = binned_column
        return df_transformed

    def _transform_spark(self, df: SparkDataFrame) -> SparkDataFrame:
        target_col = self._col + "_discretized"
        bucketizer = Bucketizer(
            splits=self._bins, inputCol=self._col, outputCol=target_col, handleInvalid=self._handle_invalid
        )
        return bucketizer.transform(df).drop(self._col).withColumnRenamed(target_col, self._col)

    def _transform_polars(self, df: PolarsDataFrame) -> PolarsDataFrame:
        warn_msg = "DataFrame will be converted to the Pandas type during internal calculations in 'transform'"
        warnings.warn(warn_msg)
        return pl.from_pandas(self._transform_pandas(df.to_pandas()))

    def transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Transforms input dataframe with fitted DiscretizingRule.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        if not self._is_fitted:
            msg = "Discretizer is not fitted"
            raise RuntimeError(msg)

        df = self._validate_input(df)

        if isinstance(df, PandasDataFrame):
            transformed_df = self._transform_pandas(df)
        elif isinstance(df, SparkDataFrame):
            transformed_df = self._transform_spark(df)
        else:
            transformed_df = self._transform_polars(df)
        return transformed_df

    def set_handle_invalid(self, handle_invalid: HandleInvalidStrategies) -> None:
        """
        Sets strategy to handle invalid values.

        :param handle_invalid: handle invalid strategy.
        """
        if handle_invalid not in self._HANDLE_INVALID_STRATEGIES:
            msg = f"handle_invalid should be either 'error' or 'skip' or 'keep', got {handle_invalid}."
            raise ValueError(msg)
        self._handle_invalid = handle_invalid

    def _validate_input(self, df: DataFrameLike) -> DataFrameLike:
        if isinstance(df, PandasDataFrame):
            df_val = df.copy()
            if (self._handle_invalid == "error") and (df_val[self._col].isna().sum() > 0):
                msg = "Data contains NaN. 'handle_invalid' param equals 'error'. \
Set 'keep' or 'skip' for processing NaN."
                raise ValueError(msg)
            if self._handle_invalid == "skip":
                df_val = df_val.dropna(subset=[self._col], axis=0)
            return df_val

        elif isinstance(df, SparkDataFrame):
            if (self._handle_invalid == "error") and (df.filter(isnan(df[self._col])).count() > 0):
                msg = "Data contains NaN. 'handle_invalid' param equals 'error'. \
Set 'keep' or 'skip' for processing NaN."
                raise ValueError(msg)
            return df

        elif isinstance(df, PolarsDataFrame):
            if (self._handle_invalid == "error") and (df[self._col].is_null().sum() > 0):
                msg = "Data contains NaN. 'handle_invalid' param equals 'error'. \
Set 'keep' or 'skip' for processing NaN."
                raise ValueError(msg)
            if self._handle_invalid == "skip":
                df = df.clone().fill_nan(None).drop_nulls(subset=[self._col])
            return df

        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)

    def save(
        self,
        path: str,
    ) -> None:
        discretizer_rule_dict = {}
        discretizer_rule_dict["_class_name"] = self.__class__.__name__
        discretizer_rule_dict["init_args"] = {
            "n_bins": self._n_bins,
            "column": self._col,
            "min_data_in_bin": self._min_data_in_bin,
            "handle_invalid": self._handle_invalid,
        }
        discretizer_rule_dict["fitted_args"] = {
            "bins": self._bins,
            "is_fitted": self._is_fitted,
        }

        base_path = Path(path).with_suffix(".replay").resolve()

        if os.path.exists(base_path):  # pragma: no cover
            msg = "There is already DiscretizingRule object saved at the given path. File will be overwrited."
            warnings.warn(msg)
        else:  # pragma: no cover
            base_path.mkdir(parents=True, exist_ok=True)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(discretizer_rule_dict, file)

    @classmethod
    def load(cls, path: str) -> "GreedyDiscretizingRule":
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            discretizer_rule_dict = json.loads(file.read())

        discretizer_rule = cls(**discretizer_rule_dict["init_args"])
        discretizer_rule._bins = discretizer_rule_dict["fitted_args"]["bins"]
        discretizer_rule._is_fitted = discretizer_rule_dict["fitted_args"]["is_fitted"]
        return discretizer_rule


class QuantileDiscretizingRule(BaseDiscretizingRule):
    """
    Implementation of the Discretizing rule for a column of PySpark, Polars and Pandas DataFrames.
    Discretizes columns values according to the Quantile strategy. All the data will be distributed
    into buckets with approximately same size.
    It is recommended to use together with the Discretizer.
    """

    _HANDLE_INVALID_STRATEGIES = ("error", "skip", "keep")

    def __init__(
        self,
        column: str,
        n_bins: int,
        handle_invalid: HandleInvalidStrategies = "keep",
    ) -> None:
        """
        :param column: Name of the column to discretize.
        :param n_bins: Number of intervals where data will be binned.
        :param handle_invalid: handle_invalid rule.
            indicates how to process NaN in data.
            If ``skip`` - filter out rows with invalid values.
            If ``error`` - throw an error.
            If ``keep`` - keep invalid values in a special additional bucket with number = n_bins.
            Default ``keep``.
        """
        self._n_bins = n_bins
        self._col = column
        self._bins = None
        self._discretizer = None
        self._is_fitted = False

        if handle_invalid not in self._HANDLE_INVALID_STRATEGIES:
            msg = f"handle_invalid should be either 'error' or 'ski[]' or 'keep', got {handle_invalid}."
            raise ValueError(msg)
        self._handle_invalid = handle_invalid

    @property
    def column(self) -> str:
        return self._col

    @property
    def n_bins(self) -> str:
        return self._n_bins

    def _fit_spark(self, df: SparkDataFrame) -> None:
        discretizer = QuantileDiscretizer(
            numBuckets=self._n_bins, inputCol=self._col, handleInvalid=self._handle_invalid
        )
        self._discretizer = discretizer.fit(df)
        self._bins = self._discretizer.getSplits()

    def _fit_pandas(self, df: PandasDataFrame) -> None:
        discretizer = KBinsDiscretizer(n_bins=self._n_bins, encode="ordinal", strategy="quantile")
        if self._handle_invalid == "keep":
            self._discretizer = discretizer.fit(df.dropna(subset=[self._col], axis=0)[[self._col]])
        else:
            self._discretizer = discretizer.fit(df[[self._col]])
        self._bins = self._discretizer.bin_edges_[0].astype(float).tolist()
        self._bins[0] = -np.inf
        self._bins[-1] = np.inf

    def _fit_polars(self, df: PolarsDataFrame) -> None:
        warn_msg = "DataFrame will be converted to the Pandas type during internal calculations in 'fit'"
        warnings.warn(warn_msg)
        self._fit_pandas(df.to_pandas())

    def fit(self, df: DataFrameLike) -> "GreedyDiscretizingRule":
        """
        Fits DiscretizingRule to input dataframe.

        :param df: input dataframe.
        :returns: fitted DiscretizingRule.
        """
        if self._is_fitted:
            return self

        df = self._validate_input(df)

        if isinstance(df, PandasDataFrame):
            self._fit_pandas(df)
        elif isinstance(df, SparkDataFrame):
            self._fit_spark(df)
        else:
            self._fit_polars(df)

        self._is_fitted = True
        return self

    def partial_fit(self, df: DataFrameLike):
        """
        Fits new data to already fitted DiscretizingRule.

        :param df: input dataframe.
        :returns: fitted DiscretizingRule.
        """
        if not self._is_fitted:
            return self.fit(df)

        msg = f"{self.__class__.__name__} is not implemented for partial_fit yet."
        raise NotImplementedError(msg)

    def _transform_pandas(self, df: PandasDataFrame) -> PandasDataFrame:
        df_nan_part = df[df[self._col].isna()]
        df_real_part = df[~df[self._col].isna()]

        binned_column = np.digitize(df_real_part[self._col].values, self._bins)
        binned_column -= 1

        df_transformed = df.copy()
        df_transformed.loc[df_real_part.index, self._col] = binned_column
        df_transformed.loc[df_nan_part.index, self._col] = [self._n_bins] * len(df_nan_part)
        return df_transformed

    def _transform_spark(self, df: SparkDataFrame) -> SparkDataFrame:
        target_col = self._col + "_discretized"
        bucketizer = Bucketizer(
            splits=self._bins, inputCol=self._col, outputCol=target_col, handleInvalid=self._handle_invalid
        )
        return bucketizer.transform(df).drop(self._col).withColumnRenamed(target_col, self._col)

    def _transform_polars(self, df: PolarsDataFrame) -> SparkDataFrame:
        warn_msg = "DataFrame will be converted to the Pandas type during internal calculations in 'transform'"
        warnings.warn(warn_msg)
        return pl.from_pandas(self._transform_pandas(df.to_pandas()))

    def transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Transforms input dataframe with fitted DiscretizingRule.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        if not self._is_fitted:
            msg = "Discretizer is not fitted"
            raise RuntimeError(msg)

        df = self._validate_input(df)

        if isinstance(df, PandasDataFrame):
            transformed_df = self._transform_pandas(df)
        elif isinstance(df, SparkDataFrame):
            transformed_df = self._transform_spark(df)
        else:
            transformed_df = self._transform_polars(df)
        return transformed_df

    def set_handle_invalid(self, handle_invalid: HandleInvalidStrategies) -> None:
        """
        Sets strategy to handle invalid values.

        :param handle_invalid: handle invalid strategy.
        """
        if handle_invalid not in self._HANDLE_INVALID_STRATEGIES:
            msg = f"handle_invalid should be either 'error' or 'skip' or 'keep', got {handle_invalid}."
            raise ValueError(msg)
        self._handle_invalid = handle_invalid

    def _validate_input(self, df: DataFrameLike) -> DataFrameLike:
        if isinstance(df, PandasDataFrame):
            if (self._handle_invalid == "error") and (df[self._col].isna().sum() > 0):
                msg = "Data contains NaN. 'handle_invalid' param equals 'error'. \
Set 'keep' or 'skip' for processing NaN."
                raise ValueError(msg)
            if self._handle_invalid == "skip":
                df = df.copy().dropna(subset=[self._col], axis=0)

        elif isinstance(df, SparkDataFrame):
            if (self._handle_invalid == "error") and (df.filter(isnan(df[self._col])).count() > 0):
                msg = "Data contains NaN. 'handle_invalid' param equals 'error'. \
Set 'keep' or 'skip' for processing NaN."
                raise ValueError(msg)
            if self._handle_invalid == "skip":
                df = df.dropna(subset=[self._col])

        elif isinstance(df, PolarsDataFrame):
            if (self._handle_invalid == "error") and (df[self._col].is_null().sum() > 0):
                msg = "Data contains NaN. 'handle_invalid' param equals 'error'. \
Set 'keep' or 'skip' for processing NaN."
                raise ValueError(msg)
            if self._handle_invalid == "skip":
                df = df.clone().fill_nan(None).drop_nulls(subset=[self._col])

        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(df)}"
            raise NotImplementedError(msg)
        return df

    def save(
        self,
        path: str,
    ) -> None:
        discretizer_rule_dict = {}
        discretizer_rule_dict["_class_name"] = self.__class__.__name__
        discretizer_rule_dict["init_args"] = {
            "n_bins": self._n_bins,
            "column": self._col,
            "handle_invalid": self._handle_invalid,
        }
        discretizer_rule_dict["fitted_args"] = {
            "bins": self._bins,
            "is_fitted": self._is_fitted,
        }

        base_path = Path(path).with_suffix(".replay").resolve()

        if os.path.exists(base_path):  # pragma: no cover
            msg = "There is already DiscretizingRule object saved at the given path. File will be overwrited."
            warnings.warn(msg)
        else:  # pragma: no cover
            base_path.mkdir(parents=True, exist_ok=True)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(discretizer_rule_dict, file)

    @classmethod
    def load(cls, path: str) -> "QuantileDiscretizingRule":
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            discretizer_rule_dict = json.loads(file.read())

        discretizer_rule = cls(**discretizer_rule_dict["init_args"])
        discretizer_rule._bins = discretizer_rule_dict["fitted_args"]["bins"]
        discretizer_rule._is_fitted = discretizer_rule_dict["fitted_args"]["is_fitted"]
        return discretizer_rule


class Discretizer:
    """
    Applies multiple discretizing rules to the data frame.
    Every sample will be distributed into bucket with number from the set [0, 1, ..., n_bins-1].
    """

    def __init__(self, rules: Sequence[BaseDiscretizingRule]):
        """
        :param rules: Sequence of rules.
        """
        self.rules = rules

    def fit(self, df: DataFrameLike) -> "Discretizer":
        """
        Fits a Discretizer by the input dataframe with given rules.

        :param df: input dataframe.
        :returns: fitted Discretizer.
        """
        for rule in self.rules:
            rule.fit(df)
        return self

    def partial_fit(self, df: DataFrameLike) -> "Discretizer":
        """
        Fits an already fitted Discretizer by the new input data frame with given rules.
        If Discretizer has not been fitted yet - performs default fit.

        :param df: input dataframe.
        :returns: fitted Discretizer.
        """
        for rule in self.rules:
            rule.partial_fit(df)
        return self

    def transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Transforms the input data frame.
        If the input data frame contains NaN values then they will be transformed by handle_invalid strategy.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        for rule in self.rules:
            df = rule.transform(df)
        return df

    def fit_transform(self, df: DataFrameLike) -> DataFrameLike:
        """
        Fits a Discretizer by the input dataframe with given rules and transforms the input dataframe.

        :param df: input dataframe.
        :returns: transformed dataframe.
        """
        return self.fit(df).transform(df)

    def set_handle_invalid(self, handle_invalid_rules: dict[str, HandleInvalidStrategies]) -> None:
        """
        Modify handle_invalid strategy on already fitted Discretizer.

        :param handle_invalid_rules: handle_invalid rule.

        Example: {"item_id" : "keep", "user_id" : "skip", "category_column" : "error"}

        Default value examples:
            If ``skip`` - filter out rows with invalid values.
            If ``error`` - throw an error.
            If ``keep`` - keep invalid values in a special additional bucket with number = n_bins.
            Default ``keep``.
        """
        columns = [i.column for i in self.rules]
        for column, handle_invalid in handle_invalid_rules.items():
            if column not in columns:
                msg = f"Column {column} not found."
                raise ValueError(msg)
            rule = list(filter(lambda x: x.column == column, self.rules))
            rule[0].set_handle_invalid(handle_invalid)

    def save(
        self,
        path: str,
    ) -> None:
        discretizer_dict = {}
        discretizer_dict["_class_name"] = self.__class__.__name__

        base_path = Path(path).with_suffix(".replay").resolve()
        if os.path.exists(base_path):  # pragma: no cover
            msg = "There is already LabelEncoder object saved at the given path. File will be overwrited."
            warnings.warn(msg)
        else:  # pragma: no cover
            base_path.mkdir(parents=True, exist_ok=True)

        discretizer_dict["rule_names"] = []

        for rule in self.rules:
            path_suffix = f"{rule.__class__.__name__}_{rule.column}"
            rule.save(str(base_path) + f"/rules/{path_suffix}")
            discretizer_dict["rule_names"].append(path_suffix)

        with open(base_path / "init_args.json", "w+") as file:
            json.dump(discretizer_dict, file)

    @classmethod
    def load(cls, path: str) -> "Discretizer":
        base_path = Path(path).with_suffix(".replay").resolve()
        with open(base_path / "init_args.json") as file:
            discretizer_dict = json.loads(file.read())
        rules = []
        for root, dirs, files in os.walk(str(base_path) + "/rules/"):
            for d in dirs:
                if d.split(".")[0] in discretizer_dict["rule_names"]:
                    with open(root + d + "/init_args.json") as file:
                        discretizer_rule_dict = json.loads(file.read())
                    rules.append(globals()[discretizer_rule_dict["_class_name"]].load(root + d))

        discretizer = cls(rules=rules)
        return discretizer



import functools
import operator
from typing import Union

import polars as pl

from replay.utils import PYSPARK_AVAILABLE, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_metric import Metric, MetricsDataFrameLike, MetricsMeanReturnType, MetricsReturnType

if PYSPARK_AVAILABLE:
    from pyspark.sql import (
        Window,
        functions as sf,
    )


class Coverage(Metric):
    """
    Metric calculation is as follows:

    * take ``K`` recommendations with the biggest ``score`` for each ``user_id``
    * count the number of distinct ``item_id`` in these recommendations
    * divide it by the number of distinct items in train dataset, provided to metric call

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> train
       query_id  item_id
    0         1        5
    1         1        6
    2         1        8
    3         1        9
    4         1        2
    5         2        5
    6         2        8
    7         2       11
    8         2        1
    9         2        3
    10        3        4
    11        3        9
    12        3        2
    >>> Coverage(2)(recommendations, train)
    {'Coverage@2': 0.5555555555555556}
    <BLANKLINE>
    """

    def __init__(
        self,
        topk: Union[list, int],
        query_column: str = "query_id",
        item_column: str = "item_id",
        rating_column: str = "rating",
        allow_caching: bool = True,
    ) -> None:
        """
        :param topk: (list or int): Consider the highest k scores in the ranking.
        :param query_column: (str): The name of the user column.
        :param item_column: (str): The name of the item column.
        :param rating_column: (str): The name of the score column.
        :param allow_caching: (bool): The flag for using caching to optimize calculations.
            Default: ``True``.
        """
        super().__init__(
            topk=topk,
            query_column=query_column,
            item_column=item_column,
            rating_column=rating_column,
        )
        self._allow_caching = allow_caching

    def _get_enriched_recommendations(
        self,
        recommendations: Union[PolarsDataFrame, SparkDataFrame],
    ) -> Union[PolarsDataFrame, SparkDataFrame]:
        if isinstance(recommendations, SparkDataFrame):
            return self._get_enriched_recommendations_spark(recommendations)
        else:
            return self._get_enriched_recommendations_polars(recommendations)

    def _get_enriched_recommendations_spark(self, recommendations: SparkDataFrame) -> SparkDataFrame:
        window = Window.partitionBy(self.query_column).orderBy(sf.col(self.rating_column).desc())
        sorted_by_score_recommendations = recommendations.withColumn("rank", sf.row_number().over(window))
        grouped_recs = (
            sorted_by_score_recommendations.select(self.item_column, "rank")
            .groupBy(self.item_column)
            .agg(sf.min("rank").alias("best_position"))
        )
        return grouped_recs

    def _get_enriched_recommendations_polars(self, recommendations: PolarsDataFrame) -> PolarsDataFrame:
        return (
            recommendations.with_columns(
                pl.col(self.rating_column)
                .rank(method="ordinal", descending=True)
                .over(self.query_column)
                .alias("__rank")
            )
            .group_by(self.item_column)
            .agg(pl.col("__rank").min().alias("best_position"))
        )

    def _spark_compute(self, recs: SparkDataFrame, train: SparkDataFrame) -> MetricsMeanReturnType:
        """
        Calculating metrics for PySpark DataFrame.
        """
        if self._allow_caching:
            recs.cache()

        item_count = train.select(self.item_column).distinct().count()

        metrics = []
        for k in self.topk:
            res = (
                recs.filter(sf.col("best_position") <= k)
                .select(self.item_column)
                .distinct()
                .join(train.select(self.item_column).distinct(), on=self.item_column)
                .count()
                / item_count
            )
            metrics.append(res)

        if self._allow_caching:
            recs.unpersist()

        return self._aggregate_results(metrics)

    def _polars_compute(self, recs: PolarsDataFrame, train: PolarsDataFrame) -> MetricsMeanReturnType:
        """
        Calculating metrics for Polars DataFrame.
        """
        item_count = train.n_unique(self.item_column)

        metrics = []
        for k in self.topk:
            res = (
                recs.filter(pl.col("best_position") <= k)
                .select(self.item_column)
                .unique()
                .join(train.select(self.item_column).unique(), on=self.item_column)
                .count()
                / item_count
            ).rows()[0][0]
            metrics.append(res)

        return self._aggregate_results(metrics)

    def _spark_call(self, recommendations: SparkDataFrame, train: SparkDataFrame) -> MetricsReturnType:
        """
        Implementation for Pyspark DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations)
        return self._spark_compute(recs, train)

    def _polars_call(self, recommendations: PolarsDataFrame, train: PolarsDataFrame) -> MetricsReturnType:
        """
        Implementation for Polars DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations)
        return self._polars_compute(recs, train)

    def _dict_call(self, recommendations: dict, train: dict) -> MetricsReturnType:
        """
        Calculating metrics in dict format.
        """
        train_items = set(functools.reduce(operator.iconcat, train.values(), []))

        len_train_items = len(train_items)
        metrics = []
        for k in self.topk:
            pred_items = set()
            for items in recommendations.values():
                for item in items[:k]:
                    pred_items.add(item)
            metrics.append(len(pred_items & train_items) / len_train_items)
        return self._aggregate_results(metrics)

    def __call__(
        self,
        recommendations: MetricsDataFrameLike,
        train: MetricsDataFrameLike,
    ) -> MetricsReturnType:
        """
        Compute metric.

        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
        :param train: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            train data.
            If DataFrame then it must contains user and item columns.
            If dict then key represents user_ids, value represents list of item_ids.

        :return: metric values
        """
        self._check_dataframes_equal_types(recommendations, train)
        if isinstance(recommendations, SparkDataFrame):
            self._check_duplicates_spark(recommendations)
            assert isinstance(train, SparkDataFrame)
            return self._spark_call(recommendations, train)
        if isinstance(recommendations, PolarsDataFrame):
            self._check_duplicates_polars(recommendations)
            assert isinstance(train, PolarsDataFrame)
            return self._polars_call(recommendations, train)
        is_pandas = isinstance(recommendations, PandasDataFrame)
        recommendations = (
            self._convert_pandas_to_dict_with_score(recommendations)
            if is_pandas
            else self._convert_dict_to_dict_with_score(recommendations)
        )
        self._check_duplicates_dict(recommendations)
        train = self._convert_pandas_to_dict_without_score(train) if is_pandas else train
        assert isinstance(train, dict)
        return self._dict_call(recommendations, train)

    @staticmethod
    def _get_metric_value_by_user(ks, *args) -> list[float]:  # pragma: no cover
        pass



from typing import Any, Literal, Optional, Protocol

import lightning
import torch
from lightning.pytorch.utilities.rank_zero import rank_zero_only

from replay.metrics.torch_metrics_builder import TorchMetricsBuilder, metrics_to_df
from replay.models.nn.sequential.postprocessors import BasePostProcessor

CallbackMetricName = Literal[
    "recall",
    "precision",
    "ndcg",
    "map",
    "mrr",
    "novelty",
    "coverage",
]


class ValidationBatch(Protocol):
    """
    Validation callback batch
    """

    query_id: torch.LongTensor
    ground_truth: torch.LongTensor
    train: torch.LongTensor


class ValidationMetricsCallback(lightning.Callback):
    """
    Callback for validation and testing stages.

    If multiple validation/testing dataloaders are used,
    the suffix of the metric name will contain the serial number of the dataloader.
    """

    def __init__(
        self,
        metrics: Optional[list[CallbackMetricName]] = None,
        ks: Optional[list[int]] = None,
        postprocessors: Optional[list[BasePostProcessor]] = None,
        item_count: Optional[int] = None,
    ):
        """
        :param metrics: Sequence of metrics to calculate.
        :param ks: highest k scores in ranking. Default: will be `[1, 5, 10, 20]`.
        :param postprocessors: postprocessors to validation stage.
        :param item_count: the total number of items in the dataset, required only for Coverage calculations.
        """
        self._metrics = metrics
        self._ks = ks
        self._item_count = item_count
        self._metrics_builders: list[TorchMetricsBuilder] = []
        self._dataloaders_size: list[int] = []
        self._postprocessors: list[BasePostProcessor] = postprocessors or []

    def _get_dataloaders_size(self, dataloaders: Optional[Any]) -> list[int]:
        if isinstance(dataloaders, torch.utils.data.DataLoader):
            return [len(dataloaders)]
        return [len(dataloader) for dataloader in dataloaders]

    def on_validation_epoch_start(
        self, trainer: lightning.Trainer, pl_module: lightning.LightningModule  # noqa: ARG002
    ) -> None:
        self._dataloaders_size = self._get_dataloaders_size(trainer.val_dataloaders)
        self._metrics_builders = [
            TorchMetricsBuilder(self._metrics, self._ks, self._item_count) for _ in self._dataloaders_size
        ]
        for builder in self._metrics_builders:
            builder.reset()

    def on_test_epoch_start(
        self,
        trainer: lightning.Trainer,
        pl_module: lightning.LightningModule,  # noqa: ARG002
    ) -> None:  # pragma: no cover
        self._dataloaders_size = self._get_dataloaders_size(trainer.test_dataloaders)
        self._metrics_builders = [
            TorchMetricsBuilder(self._metrics, self._ks, self._item_count) for _ in self._dataloaders_size
        ]
        for builder in self._metrics_builders:
            builder.reset()

    def _compute_pipeline(
        self, query_ids: torch.LongTensor, scores: torch.Tensor, ground_truth: torch.LongTensor
    ) -> tuple[torch.LongTensor, torch.Tensor, torch.LongTensor]:
        for postprocessor in self._postprocessors:
            query_ids, scores, ground_truth = postprocessor.on_validation(query_ids, scores, ground_truth)
        return query_ids, scores, ground_truth

    def on_validation_batch_end(
        self,
        trainer: lightning.Trainer,
        pl_module: lightning.LightningModule,
        outputs: torch.Tensor,
        batch: ValidationBatch,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        self._batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)

    def on_test_batch_end(
        self,
        trainer: lightning.Trainer,
        pl_module: lightning.LightningModule,
        outputs: torch.Tensor,
        batch: ValidationBatch,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:  # pragma: no cover
        self._batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)

    def _batch_end(
        self,
        trainer: lightning.Trainer,  # noqa: ARG002
        pl_module: lightning.LightningModule,
        outputs: torch.Tensor,
        batch: ValidationBatch,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        _, seen_scores, seen_ground_truth = self._compute_pipeline(batch.query_id, outputs, batch.ground_truth)
        sampled_items = torch.topk(seen_scores, k=self._metrics_builders[dataloader_idx].max_k, dim=1).indices
        self._metrics_builders[dataloader_idx].add_prediction(sampled_items, seen_ground_truth, batch.train)

        if batch_idx + 1 == self._dataloaders_size[dataloader_idx]:
            pl_module.log_dict(
                self._metrics_builders[dataloader_idx].get_metrics(),
                on_epoch=True,
                sync_dist=True,
                add_dataloader_idx=True,
            )

    def on_validation_epoch_end(self, trainer: lightning.Trainer, pl_module: lightning.LightningModule) -> None:
        self._epoch_end(trainer, pl_module)

    def on_test_epoch_end(
        self, trainer: lightning.Trainer, pl_module: lightning.LightningModule
    ) -> None:  # pragma: no cover
        self._epoch_end(trainer, pl_module)

    def _epoch_end(self, trainer: lightning.Trainer, pl_module: lightning.LightningModule) -> None:  # noqa: ARG002
        @rank_zero_only
        def print_metrics() -> None:
            metrics = {}
            for name, value in trainer.logged_metrics.items():
                if "@" in name:
                    metrics[name] = value.item()

            if metrics:
                metrics_df = metrics_to_df(metrics)

                print(metrics_df)  # noqa: T201
                print()  # noqa: T201

        print_metrics()



"""
Contains classes for users' and items' features generation based on interactions history.

``LogStatFeaturesProcessor`` to generate users' and items' features based on log.
``ConditionalPopularityProcessor`` to generate popularity among users and items
    conditioned on categorical feature value
``HistoryBasedFeaturesProcessor`` applies LogStatFeaturesProcessor
    and ConditionalPopularityProcessor as a pipeline.
"""

from datetime import datetime
from typing import Optional

from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql.types import TimestampType

    from replay.utils.spark_utils import join_or_return, join_with_col_renaming, unpersist_if_exists


class EmptyFeatureProcessor:
    """Do not perform any transformations on the dataframe"""

    def fit(self, log: SparkDataFrame, features: Optional[SparkDataFrame]) -> None:
        """
        :param log: input DataFrame ``[user_idx, item_idx, timestamp, relevance]``
        :param features: DataFrame with ``user_idx/item_idx`` and feature columns
        """

    def transform(self, log: SparkDataFrame) -> SparkDataFrame:
        """
        Return log without any transformations
        :param log: spark DataFrame
        """
        return log


class LogStatFeaturesProcessor(EmptyFeatureProcessor):
    """
    Calculate user and item features based on interactions log:
        Based on the number of interactions:
        - log number of interactions (1)
        - average log number of interactions by users interacted with item and vice versa (2)
        - difference between number of interactions by user/item (1)
        and average number of interactions (2)
        - cold user/item flag

        Based on timestamp (if present and has a TimestampType):
        - min and max interaction timestamp for user/item
        - history length (max - min timestamp)
        - log number of interactions' days
        - difference in days between last date in log and last interaction of the user/item

        Based or ratings/relevance:
        - relevance mean and std
        - relevance approximate quantiles (0.05, 0.5, 0.95)
        - abnormality of user's preferences https://hal.inria.fr/hal-01254172/document
    """

    calc_timestamp_based: bool = False
    calc_relevance_based: bool = False
    user_log_features: Optional[SparkDataFrame] = None
    item_log_features: Optional[SparkDataFrame] = None

    def _create_log_aggregates(self, agg_col: str = "user_idx") -> list:
        """
        Create features based on relevance type
        (binary or not) and whether timestamp is present.
        :param agg_col: column to create features for, user_idx or item_idx
        :return: list of columns to pass into pyspark agg
        """
        prefix = agg_col[:1]

        aggregates = [sf.log(sf.count(sf.col("relevance"))).alias(f"{prefix}_log_num_interact")]

        if self.calc_timestamp_based:
            aggregates.extend(
                [
                    sf.log(sf.countDistinct(sf.date_trunc("dd", sf.col("timestamp")))).alias(
                        f"{prefix}_log_interact_days_count"
                    ),
                    sf.min(sf.col("timestamp")).alias(f"{prefix}_min_interact_date"),
                    sf.max(sf.col("timestamp")).alias(f"{prefix}_max_interact_date"),
                ]
            )

        if self.calc_relevance_based:
            aggregates.extend(
                [
                    (
                        sf.when(
                            sf.stddev(sf.col("relevance")).isNull() | sf.isnan(sf.stddev(sf.col("relevance"))),
                            0,
                        )
                        .otherwise(sf.stddev(sf.col("relevance")))
                        .alias(f"{prefix}_std")
                    ),
                    sf.mean(sf.col("relevance")).alias(f"{prefix}_mean"),
                ]
            )
            aggregates.extend(
                sf.expr(f"percentile_approx(relevance, {percentile})").alias(f"{prefix}_quantile_{str(percentile)[2:]}")
                for percentile in [0.05, 0.5, 0.95]
            )

        return aggregates

    @staticmethod
    def _add_ts_based(features: SparkDataFrame, max_log_date: datetime, prefix: str) -> SparkDataFrame:
        """
        Add history length (max - min timestamp) and difference in days between
        last date in log and last interaction of the user/item

        :param features: dataframe with calculated log-based features
        :param max_log_date: max timestamp in log used for features calculation
        :param prefix: identifier used as a part of column name
        :return: features dataframe with new timestamp-based columns
        """
        return features.withColumn(
            f"{prefix}_history_length_days",
            sf.datediff(
                sf.col(f"{prefix}_max_interact_date"),
                sf.col(f"{prefix}_min_interact_date"),
            ),
        ).withColumn(
            f"{prefix}_last_interaction_gap_days",
            sf.datediff(sf.lit(max_log_date), sf.col(f"{prefix}_max_interact_date")),
        )

    @staticmethod
    def _cals_cross_interactions_count(log: SparkDataFrame, features: SparkDataFrame) -> SparkDataFrame:
        """
        Calculate difference between the log number of interactions by the user
        and average log number of interactions users interacted with the item has.

        :param log: dataframe with calculated log-based features
        :param features: dataframe with calculated log-based features
        :return: features dataframe with new columns
        """
        if "user_idx" in features.columns:
            new_feature_entity, calc_by_entity = "item_idx", "user_idx"
        else:
            new_feature_entity, calc_by_entity = "user_idx", "item_idx"

        mean_log_num_interact = log.join(
            features.select(calc_by_entity, f"{calc_by_entity[0]}_log_num_interact"),
            on=calc_by_entity,
            how="left",
        )
        return mean_log_num_interact.groupBy(new_feature_entity).agg(
            sf.mean(f"{calc_by_entity[0]}_log_num_interact").alias(
                f"{new_feature_entity[0]}_mean_{calc_by_entity[0]}_log_num_interact"
            )
        )

    @staticmethod
    def _calc_abnormality(log: SparkDataFrame, item_features: SparkDataFrame) -> SparkDataFrame:
        """
        Calculate  discrepancy between a rating on a resource
        and the average rating of this resource (Abnormality) and
        abnormality taking controversy of the item into account (AbnormalityCR).
        https://hal.inria.fr/hal-01254172/document

        :param log: dataframe with calculated log-based features
        :param item_features: dataframe with calculated log-based features
        :return: features dataframe with new columns
        """
        # Abnormality
        abnormality_df = join_with_col_renaming(
            left=log,
            right=item_features.select("item_idx", "i_mean", "i_std"),
            on_col_name="item_idx",
            how="left",
        )
        abnormality_df = abnormality_df.withColumn("abnormality", sf.abs(sf.col("relevance") - sf.col("i_mean")))

        abnormality_aggs = [sf.mean(sf.col("abnormality")).alias("abnormality")]

        # Abnormality CR:
        max_std = item_features.select(sf.max("i_std")).first()[0]
        min_std = item_features.select(sf.min("i_std")).first()[0]
        if max_std - min_std != 0:
            abnormality_df = abnormality_df.withColumn(
                "controversy",
                1 - (sf.col("i_std") - sf.lit(min_std)) / (sf.lit(max_std - min_std)),
            )
            abnormality_df = abnormality_df.withColumn(
                "abnormalityCR",
                (sf.col("abnormality") * sf.col("controversy")) ** 2,
            )
            abnormality_aggs.append(sf.mean(sf.col("abnormalityCR")).alias("abnormalityCR"))

        return abnormality_df.groupBy("user_idx").agg(*abnormality_aggs)

    def fit(self, log: SparkDataFrame, features: Optional[SparkDataFrame] = None) -> None:  # noqa: ARG002
        """
        Calculate log-based features for users and items

         :param log: input SparkDataFrame ``[user_idx, item_idx, timestamp, relevance]``
        """
        self.calc_timestamp_based = (isinstance(log.schema["timestamp"].dataType, TimestampType)) & (
            log.select(sf.countDistinct(sf.col("timestamp"))).first()[0] > 1
        )
        self.calc_relevance_based = log.select(sf.countDistinct(sf.col("relevance"))).first()[0] > 1

        user_log_features = log.groupBy("user_idx").agg(*self._create_log_aggregates(agg_col="user_idx"))
        item_log_features = log.groupBy("item_idx").agg(*self._create_log_aggregates(agg_col="item_idx"))

        if self.calc_timestamp_based:
            last_date = log.select(sf.max("timestamp")).first()[0]
            user_log_features = self._add_ts_based(features=user_log_features, max_log_date=last_date, prefix="u")

            item_log_features = self._add_ts_based(features=item_log_features, max_log_date=last_date, prefix="i")

        if self.calc_relevance_based:
            user_log_features = user_log_features.join(
                self._calc_abnormality(log=log, item_features=item_log_features),
                on="user_idx",
                how="left",
            ).cache()

        self.user_log_features = join_with_col_renaming(
            left=user_log_features,
            right=self._cals_cross_interactions_count(log=log, features=item_log_features),
            on_col_name="user_idx",
            how="left",
        ).cache()

        self.item_log_features = join_with_col_renaming(
            left=item_log_features,
            right=self._cals_cross_interactions_count(log=log, features=user_log_features),
            on_col_name="item_idx",
            how="left",
        ).cache()

    def transform(self, log: SparkDataFrame) -> SparkDataFrame:
        """
        Add log-based features for users and items

        :param log: input SparkDataFrame with
            ``[user_idx, item_idx, <features columns>]`` columns
        :return: log with log-based feature columns
        """
        joined = (
            log.join(
                self.user_log_features,
                on="user_idx",
                how="left",
            )
            .join(
                self.item_log_features,
                on="item_idx",
                how="left",
            )
            .withColumn(
                "na_u_log_features",
                sf.when(sf.col("u_log_num_interact").isNull(), 1.0).otherwise(0.0),
            )
            .withColumn(
                "na_i_log_features",
                sf.when(sf.col("i_log_num_interact").isNull(), 1.0).otherwise(0.0),
            )
            # TO DO std и date diff заменяем на inf, date features - будут ли работать корректно?
            # если не заменять, будет ли работать корректно?
            .fillna(dict.fromkeys(self.user_log_features.columns + self.item_log_features.columns, 0))
        )

        joined = joined.withColumn(
            "u_i_log_num_interact_diff",
            sf.col("u_log_num_interact") - sf.col("i_mean_u_log_num_interact"),
        ).withColumn(
            "i_u_log_num_interact_diff",
            sf.col("i_log_num_interact") - sf.col("u_mean_i_log_num_interact"),
        )

        return joined

    def __del__(self):
        unpersist_if_exists(self.user_log_features)
        unpersist_if_exists(self.item_log_features)


class ConditionalPopularityProcessor(EmptyFeatureProcessor):
    """
    Calculate popularity based on user or item categorical features
    (for example movie popularity among users of the same age group).
    If user features are provided, item features will be generated and vice versa.
    """

    conditional_pop_dict: Optional[dict[str, SparkDataFrame]]
    entity_name: str

    def __init__(
        self,
        cat_features_list: list,
    ):
        """
        :param cat_features_list: List of columns with categorical features to use
            for conditional popularity calculation
        """
        self.cat_features_list = cat_features_list

    def fit(self, log: SparkDataFrame, features: SparkDataFrame) -> None:
        """
        Calculate conditional popularity for id and categorical features
        defined in `cat_features_list`

        :param log: input SparkDataFrame ``[user_idx, item_idx, timestamp, relevance]``
        :param features: SparkDataFrame with ``user_idx/item_idx`` and feature columns
        """
        if len(set(self.cat_features_list).intersection(features.columns)) != len(self.cat_features_list):
            msg = (
                f"Columns {set(self.cat_features_list).difference(features.columns)} "
                f"defined in `cat_features_list` are absent in features. "
                f"features columns are: {features.columns}."
            )
            raise ValueError(msg)

        join_col, self.entity_name = (
            ("item_idx", "user_idx") if "item_idx" in features.columns else ("user_idx", "item_idx")
        )

        self.conditional_pop_dict = {}
        log_with_features = log.join(features, on=join_col, how="left")
        count_by_entity_col_name = f"count_by_{self.entity_name}"

        count_by_entity_col = log_with_features.groupBy(self.entity_name).agg(
            sf.count("relevance").alias(count_by_entity_col_name)
        )

        for cat_col in self.cat_features_list:
            col_name = f"{self.entity_name[0]}_pop_by_{cat_col}"
            intermediate_df = log_with_features.groupBy(self.entity_name, cat_col).agg(
                sf.count("relevance").alias(col_name)
            )
            intermediate_df = intermediate_df.join(
                sf.broadcast(count_by_entity_col),
                on=self.entity_name,
                how="left",
            )
            self.conditional_pop_dict[cat_col] = intermediate_df.withColumn(
                col_name, sf.col(col_name) / sf.col(count_by_entity_col_name)
            ).drop(count_by_entity_col_name)
            self.conditional_pop_dict[cat_col].cache()

    def transform(self, log: SparkDataFrame) -> SparkDataFrame:
        """
        Add conditional popularity features

        :param log: input SparkDataFrame with
            ``[user_idx, item_idx, <features columns>]`` columns
        :return: log with conditional popularity feature columns
        """

        joined = log
        for (
            key,
            value,
        ) in self.conditional_pop_dict.items():
            joined = join_or_return(
                joined,
                sf.broadcast(value),
                on=[self.entity_name, key],
                how="left",
            ).withColumn(
                f"na_{self.entity_name[0]}_pop_by_{key}",
                sf.when(
                    sf.col(f"{self.entity_name[0]}_pop_by_{key}").isNull(),
                    True,
                ).otherwise(False),
            )
            joined = joined.fillna({f"{self.entity_name[0]}_pop_by_{key}": 0})
        return joined

    def __del__(self):
        for df in self.conditional_pop_dict.values():
            unpersist_if_exists(df)


class HistoryBasedFeaturesProcessor:
    """
    Calculate user and item features based on interactions history (log).
    calculated features includes numbers of interactions, rating and timestamp distribution features
    and conditional popularity for pairs `user_idx/item_idx - categorical feature`.

    See LogStatFeaturesProcessor and ConditionalPopularityProcessor documentation
    for detailed description of generated features.
    """

    log_processor = EmptyFeatureProcessor()
    user_cond_pop_proc = EmptyFeatureProcessor()
    item_cond_pop_proc = EmptyFeatureProcessor()

    def __init__(
        self,
        use_log_features: bool = True,
        use_conditional_popularity: bool = True,
        user_cat_features_list: Optional[list] = None,
        item_cat_features_list: Optional[list] = None,
    ):
        """
        :param use_log_features: if add statistical log-based features
            generated by LogStatFeaturesProcessor
        :param use_conditional_popularity: if add conditional popularity
            features generated by ConditionalPopularityProcessor
        :param user_cat_features_list: list of user categorical features
            used to calculate item conditional popularity features
        :param item_cat_features_list: list of item categorical features
            used to calculate user conditional popularity features
        """
        if use_log_features:
            self.log_processor = LogStatFeaturesProcessor()

        if use_conditional_popularity and user_cat_features_list:
            if user_cat_features_list:
                self.user_cond_pop_proc = ConditionalPopularityProcessor(cat_features_list=user_cat_features_list)
            if item_cat_features_list:
                self.item_cond_pop_proc = ConditionalPopularityProcessor(cat_features_list=item_cat_features_list)
        self.fitted: bool = False

    def fit(
        self,
        log: SparkDataFrame,
        user_features: Optional[SparkDataFrame] = None,
        item_features: Optional[SparkDataFrame] = None,
    ) -> None:
        """
        Calculate log and conditional popularity features.

        :param log: input SparkDataFrame ``[user_idx, item_idx, timestamp, relevance]``
        :param user_features: SparkDataFrame with ``user_idx`` and feature columns
        :param item_features: SparkDataFrame with ``item_idx`` and feature columns
        """
        log = log.cache()
        self.log_processor.fit(log=log, features=user_features)
        self.user_cond_pop_proc.fit(log=log, features=user_features)
        self.item_cond_pop_proc.fit(log=log, features=item_features)
        self.fitted = True
        log.unpersist()

    def transform(
        self,
        log: SparkDataFrame,
    ):
        """
        Add features
        :param log: input SparkDataFrame with
            ``[user_idx, item_idx, <features columns>]`` columns
        :return: augmented SparkDataFrame
        """
        if not self.fitted:
            msg = "Call fit before running transform"
            raise AttributeError(msg)
        joined = self.log_processor.transform(log)
        joined = self.user_cond_pop_proc.transform(joined)
        joined = self.item_cond_pop_proc.transform(joined)

        return joined



from typing import Any, Optional

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_rec import ItemVectorModel, Recommender
from .extensions.ann.ann_mixin import ANNMixin
from .extensions.ann.index_builders.base_index_builder import IndexBuilder

if PYSPARK_AVAILABLE:
    from pyspark.ml.feature import Word2Vec
    from pyspark.ml.functions import vector_to_array
    from pyspark.ml.stat import Summarizer
    from pyspark.sql import (
        functions as sf,
        types as st,
    )

    from replay.utils.spark_utils import join_with_col_renaming, multiply_scala_udf, vector_dot


class Word2VecRec(ANNMixin, Recommender, ItemVectorModel):
    """
    Trains word2vec model where items are treated as words and queries as sentences.
    """

    def _get_ann_infer_params(self) -> dict[str, Any]:
        self.index_builder.index_params.dim = self.rank
        return {
            "features_col": "query_vector",
        }

    def _get_vectors_to_infer_ann_inner(self, interactions: SparkDataFrame, queries: SparkDataFrame) -> SparkDataFrame:
        query_vectors = self._get_query_vectors(queries, interactions)
        # converts to pandas_udf compatible format
        query_vectors = query_vectors.select(self.query_column, vector_to_array("query_vector").alias("query_vector"))
        return query_vectors

    def _configure_index_builder(self, interactions: SparkDataFrame) -> dict[str, Any]:
        item_vectors = self._get_item_vectors()
        item_vectors = item_vectors.select(self.item_column, vector_to_array("item_vector").alias("item_vector"))

        self.index_builder.index_params.dim = self.rank
        self.index_builder.index_params.max_elements = interactions.select(self.item_column).distinct().count()
        self.logger.debug("index 'num_elements' = %s", self.num_elements)
        return item_vectors, {"features_col": "item_vector", "ids_col": self.item_column}

    idf: SparkDataFrame
    vectors: SparkDataFrame

    can_predict_cold_queries = True
    _search_space = {
        "rank": {"type": "int", "args": [50, 300]},
        "window_size": {"type": "int", "args": [1, 100]},
        "use_idf": {"type": "categorical", "args": [True, False]},
    }

    def __init__(
        self,
        rank: int = 100,
        min_count: int = 5,
        step_size: int = 0.025,
        max_iter: int = 1,
        window_size: int = 1,
        use_idf: bool = False,
        seed: Optional[int] = None,
        num_partitions: Optional[int] = None,
        index_builder: Optional[IndexBuilder] = None,
    ):
        """
        :param rank: embedding size
        :param min_count: the minimum number of times a token must
            appear to be included in the word2vec model's vocabulary
        :param step_size: step size to be used for each iteration of optimization
        :param max_iter: max number of iterations
        :param window_size: window size
        :param use_idf: flag to use inverse document frequency
        :param seed: random seed
        :param index_builder: `IndexBuilder` instance that adds ANN functionality.
            If not set, then ann will not be used.
        """
        self.init_index_builder(index_builder)

        self.rank = rank
        self.window_size = window_size
        self.use_idf = use_idf
        self.min_count = min_count
        self.step_size = step_size
        self.max_iter = max_iter
        self._seed = seed
        self._num_partitions = num_partitions
        self.num_elements = None

    @property
    def _init_args(self):
        return {
            "rank": self.rank,
            "window_size": self.window_size,
            "use_idf": self.use_idf,
            "min_count": self.min_count,
            "step_size": self.step_size,
            "max_iter": self.max_iter,
            "seed": self._seed,
            "index_builder": self.index_builder.init_meta_as_dict() if self.index_builder else None,
        }

    def _save_model(self, path: str, additional_params: Optional[dict] = None):
        super()._save_model(path, additional_params)
        if self.index_builder:
            self._save_index(path)

    def _load_model(self, path: str):
        super()._load_model(path)
        if self.index_builder:
            self._load_index(path)

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        self.idf = (
            dataset.interactions.groupBy(self.item_column)
            .agg(sf.countDistinct(self.query_column).alias("count"))
            .withColumn(
                "idf",
                sf.log(sf.lit(self.queries_count) / sf.col("count")) if self.use_idf else sf.lit(1.0),
            )
            .select(self.item_column, "idf")
        )
        self.idf.cache().count()

        interactions_by_queries = (
            dataset.interactions.groupBy(self.query_column)
            .agg(sf.collect_list(sf.struct(self.timestamp_column, self.item_column)).alias("ts_item_idx"))
            .withColumn("ts_item_idx", sf.array_sort("ts_item_idx"))
            .withColumn(
                "items",
                sf.col(f"ts_item_idx.{self.item_column}").cast(st.ArrayType(st.StringType())),
            )
            .drop("ts_item_idx")
        )

        self.logger.debug("Model training")

        if self._num_partitions is None:
            self._num_partitions = interactions_by_queries.rdd.getNumPartitions()

        word_2_vec = Word2Vec(
            vectorSize=self.rank,
            minCount=self.min_count,
            numPartitions=self._num_partitions,
            stepSize=self.step_size,
            maxIter=self.max_iter,
            inputCol="items",
            outputCol="w2v_vector",
            windowSize=self.window_size,
            seed=self._seed,
        )
        self.vectors = (
            word_2_vec.fit(interactions_by_queries)
            .getVectors()
            .select(sf.col("word").cast("int").alias("item"), "vector")
        )
        self.vectors.cache().count()

    def _clear_cache(self):
        if hasattr(self, "idf") and hasattr(self, "vectors"):
            self.idf.unpersist()
            self.vectors.unpersist()

    @property
    def _dataframes(self):
        return {"idf": self.idf, "vectors": self.vectors}

    def _get_query_vectors(
        self,
        queries: SparkDataFrame,
        interactions: SparkDataFrame,
    ) -> SparkDataFrame:
        """
        :param queries: query ids, dataframe ``[query_id]``
        :param interactions: interaction dataframe
            ``[query_id, item_id, timestamp, rating]``
        :return: query embeddings dataframe
            ``[query_id, query_vector]``
        """
        res = join_with_col_renaming(interactions, queries, on_col_name=self.query_column, how="inner")
        res = join_with_col_renaming(res, self.idf, on_col_name=self.item_column, how="inner")
        res = res.join(
            self.vectors.hint("broadcast"),
            how="inner",
            on=sf.col(self.item_column) == sf.col("item"),
        ).drop("item")
        return (
            res.groupby(self.query_column)
            .agg(Summarizer.mean(multiply_scala_udf(sf.col("idf"), sf.col("vector"))).alias("query_vector"))
            .select(self.query_column, "query_vector")
        )

    def _predict_pairs_inner(
        self,
        pairs: SparkDataFrame,
        dataset: Dataset,
    ) -> SparkDataFrame:
        if dataset is None:
            msg = f"interactions is not provided, {self} predict requires interactions."
            raise ValueError(msg)

        query_vectors = self._get_query_vectors(pairs.select(self.query_column).distinct(), dataset.interactions)
        pairs_with_vectors = join_with_col_renaming(pairs, query_vectors, on_col_name=self.query_column, how="inner")
        pairs_with_vectors = pairs_with_vectors.join(
            self.vectors, on=sf.col(self.item_column) == sf.col("item"), how="inner"
        ).drop("item")
        return pairs_with_vectors.select(
            self.query_column,
            sf.col(self.item_column),
            (vector_dot(sf.col("vector"), sf.col("query_vector")) + sf.lit(self.rank)).alias(self.rating_column),
        )

    def _predict(
        self,
        dataset: Dataset,
        k: int,  # noqa: ARG002
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,  # noqa: ARG002
    ) -> SparkDataFrame:
        return self._predict_pairs_inner(queries.crossJoin(items), dataset)

    def _predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,
    ) -> SparkDataFrame:
        return self._predict_pairs_inner(pairs, dataset)

    def _get_item_vectors(self):
        return self.vectors.withColumnRenamed("vector", "item_vector").withColumnRenamed("item", self.item_column)



"""
NeighbourRec - base class that requires interactions at prediction time.
Part of set of abstract classes (from base_rec.py)
"""

from abc import ABC
from collections.abc import Iterable
from typing import Any, Optional, Union

from replay.data.dataset import Dataset
from replay.utils import PYSPARK_AVAILABLE, MissingImport, SparkDataFrame

from .base_rec import Recommender
from .extensions.ann.ann_mixin import ANNMixin

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf
    from pyspark.sql.column import Column
else:
    Column = MissingImport


class NeighbourRec(ANNMixin, Recommender, ABC):
    """Base class that requires interactions at prediction time"""

    similarity: Optional[SparkDataFrame]
    can_predict_item_to_item: bool = True
    can_predict_cold_queries: bool = True
    can_change_metric: bool = False
    item_to_item_metrics = ["similarity"]
    _similarity_metric = "similarity"

    @property
    def _dataframes(self):
        return {"similarity": self.similarity}

    def _clear_cache(self):
        if hasattr(self, "similarity"):
            self.similarity.unpersist()

    @property
    def similarity_metric(self):
        return self._similarity_metric

    @similarity_metric.setter
    def similarity_metric(self, value):
        if not self.can_change_metric:
            msg = "This class does not support changing similarity metrics"
            raise ValueError(msg)
        if value not in self.item_to_item_metrics:
            msg = f"Select one of the valid metrics for predict: {self.item_to_item_metrics}"
            raise ValueError(msg)
        self._similarity_metric = value

    def _predict_pairs_inner(
        self,
        dataset: Dataset,
        filter_df: SparkDataFrame,
        condition: Column,
        queries: SparkDataFrame,
    ) -> SparkDataFrame:
        """
        Get recommendations for all provided queries
        and filter results with ``filter_df`` by ``condition``.
        It allows to implement both ``predict_pairs`` and usual ``predict``@k.

        :param interactions: historical interactions, SparkDataFrame
            ``[user_id, item_id, timestamp, rating]``.
        :param filter_df: SparkDataFrame use to filter items:
            ``[item_idx_filter]`` or ``[user_idx_filter, item_idx_filter]``.
        :param condition: condition used for inner join with ``filter_df``
        :param queries: queries to calculate recommendations for
        :return: SparkDataFrame ``[user_id, item_id, rating]``
        """
        if dataset is None:
            msg = "interactions is not provided, but it is required for prediction"
            raise ValueError(msg)

        recs = (
            dataset.interactions.join(queries, how="inner", on=self.query_column)
            .join(
                self.similarity,
                how="inner",
                on=sf.col(self.item_column) == sf.col("item_idx_one"),
            )
            .join(
                filter_df,
                how="inner",
                on=condition,
            )
            .groupby(self.query_column, "item_idx_two")
            .agg(sf.sum(self.similarity_metric).alias(self.rating_column))
            .withColumnRenamed("item_idx_two", self.item_column)
        )
        return recs

    def _predict(
        self,
        dataset: Dataset,
        k: int,  # noqa: ARG002
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,  # noqa: ARG002
    ) -> SparkDataFrame:
        return self._predict_pairs_inner(
            dataset=dataset,
            filter_df=items.withColumnRenamed(self.item_column, "item_idx_filter"),
            condition=sf.col("item_idx_two") == sf.col("item_idx_filter"),
            queries=queries,
        )

    def _predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,
    ) -> SparkDataFrame:
        return self._predict_pairs_inner(
            dataset=dataset,
            filter_df=(
                pairs.withColumnRenamed(self.query_column, "user_idx_filter").withColumnRenamed(
                    self.item_column, "item_idx_filter"
                )
            ),
            condition=(sf.col(self.query_column) == sf.col("user_idx_filter"))
            & (sf.col("item_idx_two") == sf.col("item_idx_filter")),
            queries=pairs.select(self.query_column).distinct(),
        )

    def get_nearest_items(
        self,
        items: Union[SparkDataFrame, Iterable],
        k: int,
        metric: Optional[str] = None,
        candidates: Optional[Union[SparkDataFrame, Iterable]] = None,
    ) -> SparkDataFrame:
        """
        Get k most similar items be the `metric` for each of the `items`.

        :param items: spark dataframe or list of item ids to find neighbors
        :param k: number of neighbors
        :param metric: metric is not used to find neighbours in NeighbourRec,
            the parameter is ignored
        :param candidates: spark dataframe or list of items
            to consider as similar, e.g. popular/new items. If None,
            all items presented during model training are used.
        :return: dataframe with the most similar items an distance,
            where bigger value means greater similarity.
            spark-dataframe with columns ``[item_idx, neighbour_item_idx, similarity]``
        """

        if metric is not None:
            if metric not in self.item_to_item_metrics:
                msg = f"Select one of the valid distance metrics: {self.item_to_item_metrics}"
                raise ValueError(msg)

            self.logger.debug(
                "Metric is not used to determine nearest items in %s model",
                str(self),
            )

        return self._get_nearest_items_wrap(
            items=items,
            k=k,
            metric=metric,
            candidates=candidates,
        )

    def _get_nearest_items(
        self,
        items: SparkDataFrame,
        metric: Optional[str] = None,
        candidates: Optional[SparkDataFrame] = None,
    ) -> SparkDataFrame:
        similarity_filtered = self.similarity.join(
            items.withColumnRenamed(self.item_column, "item_idx_one"),
            on="item_idx_one",
        )

        if candidates is not None:
            similarity_filtered = similarity_filtered.join(
                candidates.withColumnRenamed(self.item_column, "item_idx_two"),
                on="item_idx_two",
            )

        return similarity_filtered.select(
            "item_idx_one",
            "item_idx_two",
            "similarity" if metric is None else metric,
        )

    def _configure_index_builder(self, interactions: SparkDataFrame) -> dict[str, Any]:
        similarity_df = self.similarity.select("similarity", "item_idx_one", "item_idx_two")
        self.index_builder.index_params.items_count = interactions.select(sf.max(self.item_column)).first()[0] + 1
        return similarity_df, {
            "features_col": None,
        }

    def _get_vectors_to_infer_ann_inner(
        self, interactions: SparkDataFrame, queries: SparkDataFrame  # noqa: ARG002
    ) -> SparkDataFrame:
        user_vectors = interactions.groupBy(self.query_column).agg(
            sf.collect_list(self.item_column).alias("vector_items"),
            sf.collect_list(self.rating_column).alias("vector_ratings"),
        )
        return user_vectors

    def _save_model(self, path: str, additional_params: Optional[dict] = None):
        super()._save_model(path, additional_params)
        if self._use_ann:
            self._save_index(path)

    def _load_model(self, path: str):
        super()._load_model(path)
        if self._use_ann:
            self._load_index(path)



"""RecSys library"""

# NOTE: This ensures distutils monkey-patching is performed before any
# functionality removed in Python 3.12 is used in downstream packages (like lightfm)
import setuptools as _

__version__ = "0.20.2"



from collections.abc import Iterable
from typing import Any, Optional, Union

from replay.data import Dataset
from replay.metrics import NDCG, Metric
from replay.models import PopRec
from replay.models.base_rec import BaseRecommender
from replay.preprocessing.filters import MinCountFilter
from replay.utils import SparkDataFrame
from replay.utils.spark_utils import fallback, get_unique_entities


class Fallback(BaseRecommender):
    """Fill missing recommendations using fallback model.
    Behaves like a recommender and have the same interface."""

    can_predict_cold_queries: bool = True

    def __init__(
        self,
        main_model: BaseRecommender,
        fallback_model: BaseRecommender = PopRec(),
        threshold: int = 0,
    ):
        """Create recommendations with `main_model`, and fill missing with `fallback_model`.
        `rating` of fallback_model will be decrease to keep main recommendations on top.

        :param main_model: initialized model
        :param fallback_model: initialized model
        :param threshold: number of interactions by which queries are divided into cold and hot
        """
        self.threshold = threshold
        self.hot_queries = None
        self.main_model = main_model
        self.fb_model = fallback_model

    # TODO: add save/load for scenarios
    @property
    def _init_args(self):
        return {"threshold": self.threshold}

    def __str__(self):
        return f"Fallback_{self.main_model!s}_{self.fb_model!s}"

    def fit(
        self,
        dataset: Dataset,
    ) -> None:
        """
        :param dataset: input Dataset with interactions and features ``[user_id, item_id, timestamp, rating]``
        :return:
        """
        query_column = dataset.feature_schema.query_id_column

        hot_data = MinCountFilter(self.threshold, query_column).transform(dataset.interactions)
        self.hot_queries = hot_data.select(query_column).distinct()
        hot_dataset = Dataset(
            feature_schema=dataset.feature_schema,
            interactions=hot_data,
            query_features=dataset.query_features,
            item_features=dataset.item_features,
            check_consistency=True,
            categorical_encoded=False,
        )
        self._fit_wrap(hot_dataset)
        self.fb_model._fit_wrap(dataset)

    def predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        """
        Get recommendations

        :param dataset: historical interactions
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: length of recommendation lists, should be less that the total number of ``items``
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``log``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``log``.
            If it contains new items, ``rating`` for them will be``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``log``.
        :return: recommendation dataframe
            ``[user_idx, item_idx, rating]``
        """
        queries = queries or dataset.interactions or dataset.query_features or self.fit_queries
        queries = get_unique_entities(queries, self.query_column)
        hot_data = MinCountFilter(self.threshold, self.query_column).transform(dataset.interactions)
        hot_queries = hot_data.select(self.query_column).distinct()
        hot_queries = hot_queries.join(queries, on=self.query_column, how="inner")

        hot_dataset = Dataset(
            feature_schema=dataset.feature_schema,
            interactions=hot_data,
            query_features=dataset.query_features,
            item_features=dataset.item_features,
            check_consistency=True,
            categorical_encoded=False,
        )

        hot_pred = self._predict_wrap(
            dataset=hot_dataset,
            k=k,
            queries=hot_queries,
            items=items,
            filter_seen_items=filter_seen_items,
        )
        cold_pred = self.fb_model._predict_wrap(
            dataset=dataset,
            k=k,
            queries=queries,
            items=items,
            filter_seen_items=filter_seen_items,
        )
        pred = fallback(hot_pred, cold_pred, k)
        return pred

    def optimize(
        self,
        train_dataset: Dataset,
        test_dataset: Dataset,
        param_borders: Optional[dict[str, dict[str, list[Any]]]] = None,
        criterion: Metric = NDCG,
        k: int = 10,
        budget: int = 10,
        new_study: bool = True,
    ) -> tuple[dict[str, Any]]:
        """
        Searches best parameters with optuna.

        :param train_dataset: train data
        :param test_dataset: test data
        :param param_borders: a dictionary with keys main and
            fallback containing dictionaries with search grid, where
            key is the parameter name and value is the range of possible values
            ``{param: [low, high]}``.
        :param criterion: metric to use for optimization
        :param k: recommendation list length
        :param budget: number of points to try
        :param new_study: keep searching with previous study or start a new study
        :return: tuple of dictionaries with best parameters
        """
        if param_borders is None:
            param_borders = {"main": None, "fallback": None}
        self.logger.info("Optimizing main model...")
        params = self.main_model.optimize(
            train_dataset,
            test_dataset,
            param_borders["main"],
            criterion,
            k,
            budget,
            new_study,
        )
        self.main_model.set_params(**params)
        if self.fb_model._search_space is not None:
            self.logger.info("Optimizing fallback model...")
            fb_params = self.fb_model.optimize(
                train_dataset,
                test_dataset,
                param_borders["fallback"],
                criterion,
                k,
                budget,
                new_study,
            )
            self.fb_model.set_params(**fb_params)
        else:
            fb_params = None
        return params, fb_params

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        self.main_model._fit_wrap(dataset)
        self.fb_model._fit_wrap(dataset)

    def _predict(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        pred = self.main_model._predict(
            dataset,
            k,
            queries,
            items,
            filter_seen_items,
        )
        return pred



import math
from typing import Any, Optional

from replay.data.dataset import Dataset
from replay.metrics import NDCG, Metric
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_rec import NonPersonalizedRecommender

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class UCB(NonPersonalizedRecommender):
    """Simple bandit model, which caclulate item rating as upper confidence bound
    (`UCB <https://medium.com/analytics-vidhya/multi-armed-bandit-analysis-of-upper-confidence-bound-algorithm-4b84be516047>`_)
    for the confidence interval of true fraction of positive ratings.
    Should be used in iterative (online) mode to achive proper recommendation quality.

    ``rating`` from interactions must be converted to binary 0-1 form.

    .. math::
        pred_i = ctr_i + \\sqrt{\\frac{c\\ln{n}}{n_i}}

    :math:`pred_i` -- predicted rating of item :math:`i`
    :math:`c` -- exploration coeficient
    :math:`n` -- number of interactions in log
    :math:`n_i` -- number of interactions with item :math:`i`

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>> data_frame = pd.DataFrame({"user_id": [1, 2, 3, 3], "item_id": [1, 2, 1, 2], "rating": [1, 0, 0, 0]})
    >>> interactions = convert2spark(data_frame)
    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="rating",
    ...             feature_type=FeatureType.NUMERICAL,
    ...             feature_hint=FeatureHint.RATING,
    ...         ),
    ...     ]
    ... )
    >>> dataset = Dataset(feature_schema, interactions)
    >>> model = UCB()
    >>> model.fit(dataset)
    >>> model.predict(dataset, k=2, queries=[1,2,3,4], items=[1,2,3]
    ... ).toPandas().sort_values(["user_id","rating","item_id"],
    ... ascending=[True,False,True]).reset_index(drop=True)
        user_id   item_id     rating
    0         1         3   2.665109
    1         1         2   1.177410
    2         2         3   2.665109
    3         2         1   1.677410
    4         3         3   2.665109
    5         4         3   2.665109
    6         4         1   1.677410

    """

    # attributes which are needed for refit method
    full_count: int
    items_counts_aggr: SparkDataFrame

    def __init__(
        self,
        exploration_coef: float = 2,
        sample: bool = False,
        seed: Optional[int] = None,
    ):
        """
        :param exploration_coef: exploration coefficient
        :param sample: flag to choose recommendation strategy.
            If True, items are sampled with a probability proportional
            to the calculated predicted rating.
            Could be changed after model training by setting the `sample` attribute.
        :param seed: random seed. Provides reproducibility if fixed
        """
        self.coef = exploration_coef
        self.sample = sample
        self.seed = seed
        super().__init__(add_cold_items=True, cold_weight=1)

    @property
    def _init_args(self):
        return {
            "exploration_coef": self.coef,
            "sample": self.sample,
            "seed": self.seed,
        }

    def optimize(
        self,
        train_dataset: Dataset,  # noqa: ARG002
        test_dataset: Dataset,  # noqa: ARG002
        param_borders: Optional[dict[str, list[Any]]] = None,  # noqa: ARG002
        criterion: Metric = NDCG,  # noqa: ARG002
        k: int = 10,  # noqa: ARG002
        budget: int = 10,  # noqa: ARG002
        new_study: bool = True,  # noqa: ARG002
    ) -> None:
        """
        Searches best parameters with optuna.

        :param train_dataset: train data
        :param test_dataset: test data
        :param param_borders: a dictionary with search borders, where
            key is the parameter name and value is the range of possible values
            ``{param: [low, high]}``. In case of categorical parameters it is
            all possible values: ``{cat_param: [cat_1, cat_2, cat_3]}``.
        :param criterion: metric to use for optimization
        :param k: recommendation list length
        :param budget: number of points to try
        :param new_study: keep searching with previous study or start a new study
        :return: dictionary with best parameters
        """
        self.logger.warning(
            "The UCB model has only exploration coefficient parameter, which cannot not be directly optimized"
        )

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        self._check_rating(dataset)

        # we save this dataframe for the refit() method
        self.items_counts_aggr = dataset.interactions.groupby(self.item_column).agg(
            sf.sum(self.rating_column).alias("pos"),
            sf.count(self.rating_column).alias("total"),
        )
        # we save this variable for the refit() method
        self.full_count = dataset.interactions.count()

        self._calc_item_popularity()

    def refit(
        self,
        dataset: Dataset,
    ) -> None:
        """Iteratively refit with new part of interactions.

        :param dataset: historical interactions with query/item features
            ``[user_id, item_id, timestamp, rating]``
        :return:
        """

        self._check_rating(dataset)

        # aggregate new interactions part
        items_counts_aggr = dataset.interactions.groupby(self.item_column).agg(
            sf.sum(self.rating_column).alias("pos"),
            sf.count(self.rating_column).alias("total"),
        )
        # combine old and new aggregations and aggregate
        self.items_counts_aggr = (
            self.items_counts_aggr.union(items_counts_aggr)
            .groupby(self.item_column)
            .agg(
                sf.sum("pos").alias("pos"),
                sf.sum("total").alias("total"),
            )
        )
        # sum old and new interactions lengths
        self.full_count += dataset.interactions.count()

        self._calc_item_popularity()

    def _calc_item_popularity(self):
        items_counts = self.items_counts_aggr.withColumn(
            self.rating_column,
            (sf.col("pos") / sf.col("total") + sf.sqrt(self.coef * sf.log(sf.lit(self.full_count)) / sf.col("total"))),
        )

        self.item_popularity = items_counts.drop("pos", "total")
        self.item_popularity.cache().count()

        self.fill = 1 + math.sqrt(self.coef * math.log(self.full_count))



import abc

import torch


class BasePostProcessor(abc.ABC):  # pragma: no cover
    """
    Abstract base class for post processor
    """

    @abc.abstractmethod
    def on_prediction(self, query_ids: torch.LongTensor, scores: torch.Tensor) -> tuple[torch.LongTensor, torch.Tensor]:
        """
        Prediction step.

        :param query_ids: query id sequence
        :param scores: calculated logits

        :returns: modified query ids and scores
        """
        # it is necessary to return the tuple of modified query_ids and scores

    @abc.abstractmethod
    def on_validation(
        self, query_ids: torch.LongTensor, scores: torch.Tensor, ground_truth: torch.LongTensor
    ) -> tuple[torch.LongTensor, torch.Tensor, torch.LongTensor]:
        """
        Validation step.

        :param query_ids: query id sequence
        :param scores: calculated logits
        :param ground_truth: ground truth dataset

        :returns: modified query ids and scores and ground truth dataset
        """
        # it is necessary to return the tuple of modified query_ids, scores and ground_truth



import warnings
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, Optional, Union

import numpy as np
import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .descriptors import CalculationDescriptor, Mean

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf
    from pyspark.sql.types import ArrayType, DoubleType, StructType


MetricsDataFrameLike = Union[DataFrameLike, dict]
MetricsMeanReturnType = Mapping[str, float]
MetricsPerUserReturnType = Mapping[str, Mapping[Any, float]]
MetricsReturnType = Union[MetricsMeanReturnType, MetricsPerUserReturnType]


class MetricDuplicatesWarning(Warning):
    """Recommendations contain duplicates"""


class Metric(ABC):
    """Base metric class"""

    def __init__(
        self,
        topk: Union[list[int], int],
        query_column: str = "query_id",
        item_column: str = "item_id",
        rating_column: str = "rating",
        mode: CalculationDescriptor = Mean(),
    ) -> None:
        """
        :param topk: (list or int): Consider the highest k scores in the ranking.
        :param query_column: (str): The name of the user column.
        :param item_column: (str): The name of the item column.
        :param rating_column: (str): The name of the score column.
        :param mode: (CalculationDescriptor): class for calculating aggregation metrics.
            Default: ``Mean``.
        """
        if isinstance(topk, list):
            for item in topk:
                if not isinstance(item, int):
                    msg = f"{item} is not int"
                    raise ValueError(msg)
        elif isinstance(topk, int):
            topk = [topk]
        else:
            msg = "topk not list or int"
            raise ValueError(msg)
        self.topk = sorted(topk)
        self.query_column = query_column
        self.item_column = item_column
        self.rating_column = rating_column
        self._mode = mode

    @property
    def __name__(self) -> str:
        mode_name = self._mode.__name__
        return str(type(self).__name__) + (f"-{mode_name}" if mode_name != "Mean" else "")

    def _check_dataframes_equal_types(
        self,
        recommendations: MetricsDataFrameLike,
        ground_truth: MetricsDataFrameLike,
    ) -> None:
        """
        Types of all data frames must be the same.
        """
        if not isinstance(recommendations, type(ground_truth)):
            msg = "All given data frames must have the same type"
            raise ValueError(msg)

    def _duplicate_warn(self):
        warnings.warn(
            "The recommendations contain duplicated users and items.The metrics may be higher than the actual ones.",
            MetricDuplicatesWarning,
        )

    def _check_duplicates_spark(self, recommendations: SparkDataFrame) -> None:
        duplicates_count = (
            recommendations.groupBy(self.query_column, self.item_column).count().filter("count >= 2").count()
        )
        if duplicates_count:
            self._duplicate_warn()

    def _check_duplicates_dict(self, recommendations: dict) -> None:
        for items in recommendations.values():
            items_set = set(items)
            if len(items) != len(items_set):
                self._duplicate_warn()
                return

    def _check_duplicates_polars(self, recommendations: PolarsDataFrame) -> None:
        duplicates_count = recommendations.group_by(self.query_column, self.item_column).len().filter(pl.col("len") > 1)
        if not duplicates_count.is_empty():
            self._duplicate_warn()

    def __call__(
        self,
        recommendations: MetricsDataFrameLike,
        ground_truth: MetricsDataFrameLike,
    ) -> MetricsReturnType:
        """
        Compute metric.

        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict): model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
        :param ground_truth: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            test data.
            If DataFrame then it must contains user and item columns.
            If dict then key represents user_ids, value represents list of item_ids.

        :return: metric values
        """
        self._check_dataframes_equal_types(recommendations, ground_truth)
        if isinstance(recommendations, SparkDataFrame):
            self._check_duplicates_spark(recommendations)
            assert isinstance(ground_truth, SparkDataFrame)
            return self._spark_call(recommendations, ground_truth)
        if isinstance(recommendations, PolarsDataFrame):
            self._check_duplicates_polars(recommendations)
            assert isinstance(ground_truth, PolarsDataFrame)
            return self._polars_call(recommendations, ground_truth)
        is_pandas = isinstance(recommendations, PandasDataFrame)
        recommendations = (
            self._convert_pandas_to_dict_with_score(recommendations)
            if is_pandas
            else self._convert_dict_to_dict_with_score(recommendations)
        )
        self._check_duplicates_dict(recommendations)
        ground_truth = self._convert_pandas_to_dict_without_score(ground_truth) if is_pandas else ground_truth
        assert isinstance(ground_truth, dict)
        return self._dict_call(
            list(ground_truth),
            pred_item_id=recommendations,
            ground_truth=ground_truth,
        )

    def _convert_pandas_to_dict_with_score(self, data: PandasDataFrame) -> dict:
        return (
            data.sort_values(by=[self.rating_column, self.item_column], ascending=False, kind="stable")
            .groupby(self.query_column)[self.item_column]
            .apply(list)
            .to_dict()
        )

    def _convert_dict_to_dict_with_score(self, data: dict) -> dict:
        converted_data = {}
        for user, items in data.items():
            is_sorted = True
            for i in range(1, len(items)):
                is_sorted &= items[i - 1][1] >= items[i][1]
                if not is_sorted:
                    break
            if not is_sorted:
                items = sorted(items, key=lambda x: x[1], reverse=True)
            converted_data[user] = [item for item, _ in items]
        return converted_data

    def _convert_pandas_to_dict_without_score(self, data: PandasDataFrame) -> dict:
        return data.groupby(self.query_column)[self.item_column].apply(list).to_dict()

    def _dict_call(self, users: list, **kwargs: dict) -> MetricsReturnType:
        """
        Calculating metrics in dict format.
        kwargs can contain different dicts (for example, ground_truth or train), it depends on the metric.
        """

        keys_list = sorted(kwargs.keys())
        distribution_per_user = {}
        for user in users:
            args = [kwargs[key].get(user, None) for key in keys_list]
            distribution_per_user[user] = self._get_metric_value_by_user(self.topk, *args)
        if self._mode.__name__ == "PerUser":
            return self._aggregate_results_per_user(distribution_per_user)
        distribution = np.stack(list(distribution_per_user.values()))
        assert distribution.shape[1] == len(self.topk)
        metrics = [self._mode.cpu(distribution[:, k]) for k in range(distribution.shape[1])]
        return self._aggregate_results(metrics)

    def _get_items_list_per_user_spark(
        self, recommendations: SparkDataFrame, extra_column: Optional[str] = None
    ) -> SparkDataFrame:
        recommendations = recommendations.groupby(self.query_column).agg(
            sf.sort_array(
                sf.collect_list(
                    sf.struct(*[c for c in [self.rating_column, self.item_column, extra_column] if c is not None])
                ),
                False,
            ).alias("pred")
        )
        selection = [
            self.query_column,
            sf.col(f"pred.{self.item_column}").alias("pred_item_id"),
        ]
        if extra_column:
            selection.append(sf.col(f"pred.{extra_column}").alias(extra_column))

        recommendations = recommendations.select(*selection)
        return recommendations

    def _get_items_list_per_user_polars(
        self, recommendations: PolarsDataFrame, extra_column: Optional[str] = None
    ) -> PolarsDataFrame:
        selection = [self.query_column, "pred_item_id"]
        sorting = [self.rating_column, self.item_column]
        agg = [pl.col(self.item_column)]
        if extra_column is not None:
            sorting.append(extra_column)
            agg.append(pl.col(extra_column))
            selection.append(extra_column)

        recommendations = (
            recommendations.sort(sorting, descending=True)
            .group_by(self.query_column)
            .agg(*agg)
            .rename({self.item_column: "pred_item_id"})
        )

        recommendations = recommendations.select(*selection)
        return recommendations

    def _get_items_list_per_user(
        self, recommendations: Union[SparkDataFrame, PolarsDataFrame], extra_column: Optional[str] = None
    ) -> Union[SparkDataFrame, PolarsDataFrame]:
        if isinstance(recommendations, SparkDataFrame):
            return self._get_items_list_per_user_spark(recommendations, extra_column)
        else:
            return self._get_items_list_per_user_polars(recommendations, extra_column)

    def _rearrange_columns(
        self, data: Union[SparkDataFrame, PolarsDataFrame]
    ) -> Union[SparkDataFrame, PolarsDataFrame]:
        cols = data.columns
        cols.remove(self.query_column)
        cols = [self.query_column, *sorted(cols)]
        return data.select(*cols)

    def _get_enriched_recommendations(
        self,
        recommendations: Union[PolarsDataFrame, SparkDataFrame],
        ground_truth: Union[PolarsDataFrame, SparkDataFrame],
    ) -> Union[PolarsDataFrame, SparkDataFrame]:
        if isinstance(recommendations, SparkDataFrame):
            return self._get_enriched_recommendations_spark(recommendations, ground_truth)
        else:
            return self._get_enriched_recommendations_polars(recommendations, ground_truth)

    def _get_enriched_recommendations_spark(
        self,
        recommendations: SparkDataFrame,
        ground_truth: SparkDataFrame,
    ) -> SparkDataFrame:
        true_items_by_users = ground_truth.groupby(self.query_column).agg(
            sf.collect_set(self.item_column).alias("ground_truth")
        )

        sorted_by_score_recommendations = self._get_items_list_per_user(recommendations)

        enriched_recommendations = sorted_by_score_recommendations.join(
            true_items_by_users, on=self.query_column, how="right"
        )
        return self._rearrange_columns(enriched_recommendations)

    def _get_enriched_recommendations_polars(
        self,
        recommendations: PolarsDataFrame,
        ground_truth: PolarsDataFrame,
    ) -> PolarsDataFrame:
        true_items_by_users = (
            ground_truth.group_by(self.query_column)
            .agg(pl.col(self.item_column))
            .rename({self.item_column: "ground_truth"})
        )

        sorted_by_score_recommendations = self._get_items_list_per_user(recommendations)

        enriched_recommendations = true_items_by_users.join(
            sorted_by_score_recommendations, on=self.query_column, how="left"
        )
        return self._rearrange_columns(enriched_recommendations)

    def _aggregate_results_per_user(self, distribution_per_user: dict[Any, list[float]]) -> MetricsPerUserReturnType:
        res: MetricsPerUserReturnType = {}
        for index, val in enumerate(self.topk):
            metric_name = f"{self.__name__}@{val}"
            res[metric_name] = {}
            for user, metrics in distribution_per_user.items():
                res[metric_name][user] = metrics[index]
        return res

    def _aggregate_results(self, metrics: list) -> MetricsMeanReturnType:
        res = {}
        for index, val in enumerate(self.topk):
            metric_name = f"{self.__name__}@{val}"
            res[metric_name] = metrics[index]
        return res

    def _spark_compute(self, recs: SparkDataFrame) -> MetricsReturnType:
        """
        Calculating metrics for PySpark DataFrame.
        """
        recs_with_topk_list = recs.withColumn("k", sf.array(*[sf.lit(x) for x in self.topk]))
        distribution = self._get_metric_distribution(recs_with_topk_list)
        if self._mode.__name__ == "PerUser":
            return self._aggregate_results_per_user(distribution.rdd.collectAsMap())
        metrics = [
            self._mode.spark(distribution.select(sf.col("value").getItem(i)).withColumnRenamed(f"value[{i}]", "val"))
            for i in range(len(self.topk))
        ]
        return self._aggregate_results(metrics)

    def _polars_compute(self, recs: PolarsDataFrame) -> MetricsReturnType:
        distribution = self._get_metric_distribution(recs)
        if self._mode.__name__ == "PerUser":
            return self._aggregate_results_per_user(
                dict(
                    distribution.select(
                        self.query_column, value=pl.concat_list(pl.exclude(self.query_column))
                    ).iter_rows()
                )
            )
        metrics = [self._mode.cpu(distribution.select(column)) for column in distribution.columns[1:]]
        return self._aggregate_results(metrics)

    def _spark_call(self, recommendations: SparkDataFrame, ground_truth: SparkDataFrame) -> MetricsReturnType:
        """
        Implementation for PySpark DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations, ground_truth)
        return self._spark_compute(recs)

    def _polars_call(self, recommendations: PolarsDataFrame, ground_truth: PolarsDataFrame) -> MetricsReturnType:
        """
        Implementation for Polars DataFrame.
        """
        recs = self._get_enriched_recommendations(recommendations, ground_truth)
        return self._polars_compute(recs)

    def _get_metric_distribution(
        self, recs: Union[PolarsDataFrame, SparkDataFrame]
    ) -> Union[PolarsDataFrame, SparkDataFrame]:
        if isinstance(recs, SparkDataFrame):
            return self._get_metric_distribution_spark(recs)
        else:
            return self._get_metric_distribution_polars(recs)

    def _get_metric_distribution_spark(self, recs: SparkDataFrame) -> SparkDataFrame:
        cur_class = self.__class__
        distribution = recs.rdd.flatMap(  # pragma: no cover, due to incorrect work of coverage tool
            lambda x: [(x[0], cur_class._get_metric_value_by_user(x[-1], *x[1:-1]))]
        ).toDF(
            StructType()
            .add("user_id", recs.schema[self.query_column].dataType.typeName(), False)
            .add("value", ArrayType(DoubleType()), False)
        )
        return distribution

    def _get_metric_distribution_polars(self, recs: PolarsDataFrame) -> PolarsDataFrame:
        cur_class = self.__class__
        distribution = recs.map_rows(lambda x: (x[0], *cur_class._get_metric_value_by_user(self.topk, *x[1:])))
        distribution = distribution.rename({"column_0": self.query_column})
        distribution = distribution.rename(
            {distribution.columns[x + 1]: f"value_{self.topk[x]}" for x in range(len(self.topk))}
        )
        return distribution

    @staticmethod
    @abstractmethod
    def _get_metric_value_by_user(ks: list[int], *args: list) -> list[float]:  # pragma: no cover
        """
        Metric calculation for one user.

        :param k: depth cut-off
        :param ground_truth: test data
        :param pred: recommendations
        :return: metric value for current user
        """
        raise NotImplementedError()



from collections import defaultdict
from typing import Union

import numpy as np
import polars as pl

from replay.utils import PYSPARK_AVAILABLE, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_metric import Metric, MetricsDataFrameLike, MetricsReturnType

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class Surprisal(Metric):
    """
    Measures how many surprising rare items are present in recommendations.

    .. math::
        \\textit{Self-Information}(j)= -\\log_2 \\frac {u_j}{N}

    :math:`u_j` -- number of users that interacted with item :math:`j`.
    Cold items are treated as if they were rated by 1 user.
    That is, if they appear in recommendations it will be completely unexpected.

    Surprisal for item :math:`j` is

    .. math::
        Surprisal(j)= \\frac {\\textit{Self-Information}(j)}{log_2 N}

    Recommendation list surprisal is the average surprisal of items in it.

    .. math::
        Surprisal@K(i) = \\frac {\\sum_{j=1}^{K}Surprisal(j)} {K}

    Final metric is averaged by users.

    .. math::
        Surprisal@K = \\frac {\\sum_{i=1}^{N}Surprisal@K(i)}{N}

    :math:`N` -- the number of users.

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> train
       query_id  item_id
    0         1        5
    1         1        6
    2         1        8
    3         1        9
    4         1        2
    5         2        5
    6         2        8
    7         2       11
    8         2        1
    9         2        3
    10        3        4
    11        3        9
    12        3        2
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> Surprisal(2)(recommendations, train)
    {'Surprisal@2': 0.6845351232142715}
    >>> Surprisal(2, mode=PerUser())(recommendations, train)
    {'Surprisal-PerUser@2': {1: 1.0, 2: 0.3690702464285426, 3: 0.6845351232142713}}
    >>> Surprisal(2, mode=Median())(recommendations, train)
    {'Surprisal-Median@2': 0.6845351232142713}
    >>> Surprisal(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, train)
    {'Surprisal-ConfidenceInterval@2': 0.3569755541728279}
    <BLANKLINE>
    """

    def _get_weights(self, train: dict) -> dict:
        n_users = len(train.keys())
        items_counter = defaultdict(set)
        for user, items in train.items():
            for item in items:
                items_counter[item].add(user)
        weights = {}
        for item, users in items_counter.items():
            weights[item] = np.log2(n_users / len(users)) / np.log2(n_users)
        return weights

    def _get_recommendation_weights(self, recommendations: dict, train: dict) -> dict:
        weights = self._get_weights(train)
        recs_with_weights = {}
        for user, items in recommendations.items():
            recs_with_weights[user] = [weights.get(i, 1) for i in items]
        return recs_with_weights

    def _get_enriched_recommendations(
        self,
        recommendations: Union[PolarsDataFrame, SparkDataFrame],
        train: Union[PolarsDataFrame, SparkDataFrame],
    ) -> Union[PolarsDataFrame, SparkDataFrame]:
        if isinstance(recommendations, SparkDataFrame):
            return self._get_enriched_recommendations_spark(recommendations, train)
        else:
            return self._get_enriched_recommendations_polars(recommendations, train)

    def _get_enriched_recommendations_spark(
        self, recommendations: SparkDataFrame, train: SparkDataFrame
    ) -> SparkDataFrame:
        n_users = train.select(self.query_column).distinct().count()
        item_weights = train.groupby(self.item_column).agg(
            (sf.log2(n_users / sf.countDistinct(self.query_column)) / np.log2(n_users)).alias("weight")
        )
        recommendations = recommendations.join(item_weights, on=self.item_column, how="left").fillna(1.0)

        sorted_by_score_recommendations = self._get_items_list_per_user(recommendations, "weight")
        return self._rearrange_columns(sorted_by_score_recommendations)

    def _get_enriched_recommendations_polars(
        self, recommendations: PolarsDataFrame, train: PolarsDataFrame
    ) -> PolarsDataFrame:
        n_users = train.select(self.query_column).n_unique()
        item_weights = train.group_by(self.item_column).agg(
            (np.log2(n_users / pl.col(self.query_column).n_unique()) / np.log2(n_users)).alias("weight")
        )
        recommendations = recommendations.join(item_weights, on=self.item_column, how="left").with_columns(
            pl.col("weight").fill_null(1.0)
        )

        sorted_by_score_recommendations = self._get_items_list_per_user(recommendations, "weight")
        return self._rearrange_columns(sorted_by_score_recommendations)

    def __call__(
        self,
        recommendations: MetricsDataFrameLike,
        train: MetricsDataFrameLike,
    ) -> MetricsReturnType:
        """
        Compute metric.

        Args:
            recommendations (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
                model predictions.
                If DataFrame then it must contains user, item and score columns.
                If dict then items must be sorted in decreasing order of their scores.
            train (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict, optional):
                train data.
                If DataFrame then it must contains user and item columns.

        Returns:
            dict: metric values
        """
        self._check_dataframes_equal_types(recommendations, train)
        if isinstance(recommendations, SparkDataFrame):
            self._check_duplicates_spark(recommendations)
            assert isinstance(train, SparkDataFrame)
            return self._spark_call(recommendations, train)
        if isinstance(recommendations, PolarsDataFrame):
            self._check_duplicates_polars(recommendations)
            assert isinstance(train, PolarsDataFrame)
            return self._polars_call(recommendations, train)
        is_pandas = isinstance(recommendations, PandasDataFrame)
        recommendations = (
            self._convert_pandas_to_dict_with_score(recommendations)
            if is_pandas
            else self._convert_dict_to_dict_with_score(recommendations)
        )
        self._check_duplicates_dict(recommendations)
        train = self._convert_pandas_to_dict_without_score(train) if is_pandas else train
        assert isinstance(train, dict)

        weights = self._get_recommendation_weights(recommendations, train)
        return self._dict_call(
            list(recommendations),
            pred_item_id=recommendations,
            pred_weight=weights,
        )

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], pred_item_ids: list, pred_weights: list) -> list[float]:
        if not pred_item_ids:
            return [0.0 for _ in ks]
        res = []
        for k in ks:
            ans = 0
            for weight, _ in zip(pred_weights[:k], pred_item_ids[:k]):
                ans += weight
            res.append(ans / k)
        return res
